use std::{fmt, path::PathBuf};

use serde::{Deserialize, Serialize};
use tokio::{
    fs::OpenOptions,
    io::{AsyncReadExt, AsyncWriteExt, BufReader},
};

const CONFIG_FILE_NAME: &str = "saved_devices.json";

#[derive(Clone, Deserialize, Serialize, Debug)]
pub struct Device {
    name: String,
    id: String,
    user: String,
    device: String,
}

impl fmt::Display for Device {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        self.name.fmt(f)
    }
}

/// Used by /api/v0/request_device
#[derive(Clone, Deserialize, Serialize, Debug)]
pub struct SimplifiedDevice {
    user: String,
    device: String,
}

impl Device {
    pub fn simplify(&self) -> SimplifiedDevice {
        SimplifiedDevice {
            user: self.user.clone(),
            device: self.device.clone(),
        }
    }
}

pub enum Error {
    IoError(std::io::Error),
    JsonError(serde_json::Error),
}

impl fmt::Display for Error {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        match self {
            Self::IoError(err) => err.fmt(f),
            Self::JsonError(err) => err.fmt(f),
        }
    }
}

impl From<std::io::Error> for Error {
    fn from(value: std::io::Error) -> Self {
        Self::IoError(value)
    }
}

impl From<serde_json::Error> for Error {
    fn from(value: serde_json::Error) -> Self {
        Self::JsonError(value)
    }
}

pub struct DeviceList {
    devices: Vec<Device>,
    modified: bool,
}

impl std::ops::Deref for DeviceList {
    type Target = Vec<Device>;
    fn deref(&self) -> &Self::Target {
        &self.devices
    }
}

impl DeviceList {
    pub async fn load() -> Result<Self, Error> {
        let path = cache_path();
        let devices: Vec<Device> = if path.exists() {
            let mut buf: Vec<u8> = Vec::new();

            let file = OpenOptions::new().read(true).open(&path).await?;
            let mut reader = BufReader::new(file);
            reader.read_to_end(&mut buf).await?;
            serde_json::from_slice(&buf)?
        } else {
            Vec::new()
        };

        Ok(Self {
            devices,
            modified: false,
        })
    }

    pub fn add(&mut self, dev: Device) {
        if let Some(pos) = self.devices.iter().position(|x| x.id == dev.id) {
            self.devices.remove(pos);
        }
        self.devices.push(dev);
        self.modified = true;
    }

    pub fn drop_by_name(&mut self, name: String) {
        if let Some(pos) = self.devices.iter().position(|x| x.name == name) {
            self.devices.remove(pos);
            self.modified = true;
        }
    }

    pub fn find(&self, name: impl AsRef<str>) -> Option<&Device> {
        self.devices.iter().find(|x| x.name == name.as_ref())
    }

    pub fn find_id(&self, id: impl AsRef<str>) -> Option<&Device> {
        self.devices.iter().find(|x| x.id == id.as_ref())
    }

    pub async fn commit(&mut self) -> Result<(), Error> {
        if self.modified {
            let path = cache_path();
            let parent = path.parent().unwrap();
            if !parent.exists() {
                tokio::fs::create_dir_all(parent).await?;
            }
            let mut file = OpenOptions::new()
                .write(true)
                .create(true)
                .truncate(true)
                .open(path)
                .await?;
            let vec_data = serde_json::to_vec(&self.devices)?;
            file.write_all(&vec_data).await?;

            self.modified = false;
        }

        Ok(())
    }
}

#[inline]
fn cache_path() -> PathBuf {
    dirs::config_dir()
        .unwrap()
        .join(crate::CONFIG_DIR_NAME)
        .join(CONFIG_FILE_NAME)
}
